#include <iostream>
#include <string>
#include <thread>
#include <mutex>
#include <deque>
#include <memory>
#include <condition_variable>
#include <cstdlib>

using namespace std;

const unsigned int stack_size = 1000;

class SimpleStack
{
  int stack[stack_size];
  unsigned int length;

public:
  SimpleStack() : length(0)
  {}

  bool push(const int & e)
  {
    if (length < stack_size)
    {
      stack[length++] = e;
      return true;
    }

    return false;
  }

  bool pop(int & e)
  {
    if (length > 0)
    {
      e = stack[--length];
      return true;
    }

    return false;
  }

  unsigned int size()
  {
    return length;
  }
};

//thread-safe version of stack
class MonitorObject
{
  SimpleStack stack;

  //synchronization
  mutex lock;
  condition_variable not_full;
  condition_variable not_empty;

public:
  void push(const int & e)
  {
    unique_lock<mutex> l(lock);
    not_full.wait(l, [&](){ return (stack_size != stack.size()); });

    stack.push(e);

    not_empty.notify_one();
  }

  void pop(int & e)
  {
    unique_lock<mutex> l(lock);
    not_empty.wait(l, [&](){ return (0 != stack.size()); });

    stack.pop(e);

    not_full.notify_one();
  }

  unsigned int size()
  {
    lock_guard<mutex> l(lock);
    return stack.size();
  }
};

mutex print_mtx;

void consumer(string name, MonitorObject & mo)
{
  for (int i = 0; i < 20; ++i)
  {
    int value;
    mo.pop(value);
    
    print_mtx.lock();
    cout << "Consumer " << name << " pop value: " << value << endl;
    print_mtx.unlock();
    
    this_thread::sleep_for(chrono::milliseconds(500));
  }
}

void producer(string name, MonitorObject & mo)
{
  for (int i = 0; i < 20; ++i)
  {
    int value = rand();
    mo.push(value);

    print_mtx.lock();
    cout << "Producer " << name << " push value: " << value << endl;
    print_mtx.unlock();

    this_thread::sleep_for(chrono::milliseconds(250));
  }
}

int main()
{
  MonitorObject stack;

  thread p(producer, "PRDR", ref(stack));
  thread c(consumer, "CNSR", ref(stack));

  p.join();
  c.join();

  return 0;
}